#include "AneuMeshLoader.h"
#include <ranges>
#include <fstream>
#include <cmath>
using namespace std;
//!!! Задание на защиту (не удалять!)
//!!! Реализовать на основе std::ranges
//!!! 1. Создать контейнер тетраэдральных КЭ из данной области.
//!!! 2. С помощью потокового итератора вывести в файл списки ID узлов тех тетраэдров,
//!!!    у которых минимальный двугранный угол от 0 до 30 градусов. 

struct ABCD {
	double A = 0, B = 0, C = 0, D = 0;
	ABCD(Node& a, Node& b, Node& c) {
		double x1 = a.Coordin.at(0), y1 = a.Coordin.at(1), z1 = a.Coordin.at(2),
			   x2 = b.Coordin.at(0), y2 = b.Coordin.at(1), z2 = b.Coordin.at(2),
			   x3 = c.Coordin.at(0), y3 = c.Coordin.at(1), z3 = c.Coordin.at(2);
		A = y1 * z2 - y1 * z3 - y2 * z1 + y2 * z3 + y3 * z1 - y3 * z2;
		B = -x1 * z2 + x1 * z3 + x2 * z1 - x2 * z3 - x3 * z1 + x3 * z2;
		C = x1 * y2 - x1 * y3 - x2 * y1 + x2 * y3 + x3 * y1 - x3 * y2;
		D = -x1 * y2 * z3 + x1 * y3 * z2 + x2 * y1 * z3 - x2 * y3 * z1 - x3 * y1 * z2 + x3 * y2 * z1;
	};
};

bool function(AneuMeshLoader& MyKE,int& id1_edge, int& id2_edge, int& id1, int& id2) {
	ABCD plane1(MyKE.vector_Node().at(id1_edge-1), MyKE.vector_Node().at(id2_edge-1), MyKE.vector_Node().at(id1-1));
	ABCD plane2(MyKE.vector_Node().at(id1_edge-1), MyKE.vector_Node().at(id2_edge-1), MyKE.vector_Node().at(id2-1));
	double cos = (plane1.A * plane2.A + plane1.B * plane2.B + plane1.C * plane2.C) /
		(sqrt(plane1.A * plane1.A + plane1.B * plane1.B + plane1.C * plane1.C) * sqrt(plane2.A * plane2.A + plane2.B * plane2.B + plane2.C * plane2.C));
	if (cos < 1 / 2 && cos>0) return true;
	else return false;
}

void v(AneuMeshLoader& MyKE) {
	int id;
	cout << "Enter border ID:";
	cin >> id;
	std::set<FiniteElement> Container=*MyKE.ContainerKE_by_id_border(id);
	ofstream file("MyFile.txt");
	if (!file.is_open()) cout << "Error";
	else {
		auto selection = [&](auto& elem) {
			auto& el = elem.ID_Nodes;
			return (function(MyKE, el.at(0), el.at(1), el.at(2), el.at(3)) ||
				function(MyKE, el.at(0), el.at(2), el.at(1), el.at(3)) ||
				function(MyKE, el.at(0), el.at(3), el.at(2), el.at(1)) ||
				function(MyKE, el.at(1), el.at(2), el.at(0), el.at(3)) ||
				function(MyKE, el.at(1), el.at(3), el.at(2), el.at(0)) ||
				function(MyKE, el.at(2), el.at(3), el.at(0), el.at(1)));
					};
		auto alg = Container | std::ranges::views::filter(selection);
		for (auto i : alg) {
			file >> alg.ID_Nodes;
		}
	}

}

//!!! Это куда-то в Data_types.h


int main() {
	AneuMeshLoader MyKE; //!!! Утечка памяти
	
	
	std::string str;
	cout << "Enter the file name:";
	cin >> str;
	MyKE.loadMesh(str);

	cout << "---------------STL-container nodes---------------"<<endl;
	cout << MyKE.vector_Node() <<endl;

	cout << "---------------STL-container finite elements---------------"<<endl;
	cout << MyKE.vector_FiniteElement()<<endl;

	cout << "---------------STL-container boundary finite elements---------------"<<endl;
	cout << MyKE.vector_BoundaryFiniteElement() <<endl;

	cout << "---------------Finite element by three nodes---------------" << endl;
	cout << "Enter three nodes:";
	int id1, id2, id3;
	cin >> id1 >> id2 >> id3;
	cout << *MyKE.FiniteElementByThreeNodes(id1,id2,id3) << endl;

	cout << "---------------Finite element by two nodes---------------" << endl;
	cout << "Enter two nodes:";
	cin >> id1 >> id2;
	cout << *MyKE.FiniteElementByTwoNodes(id1, id2) << endl;

	cout << "---------------Container of boundary nodes by border ID---------------" << endl;
	cout << "Enter border ID:";
	cin >> id1;
	cout << *MyKE.BoundaryNodes(id1) << endl;

	cout << "---------------Container KE by border ID---------------" << endl;
	cout << "Enter border ID:";
	cin >> id1;
	cout << *MyKE.ContainerKE_by_id_border(id1) << endl;

	cout << "---------------Container boundary KE by border ID---------------" << endl;
	cout << "Enter border ID:";
	cin >> id1;
	cout << *MyKE.BoundaryKE(id1) << endl;

	cout << "---------------Container with containers of \"neighboring\" nodes---------------" << endl;
	cout << *MyKE.Container_of_containers_of_neighboring_nodes() << endl;
	
	cout << "---------------Print Node---------------" << endl;
	MyKE.print_Node();

	cout << "---------------Print FiniteElement---------------" << endl;
	MyKE.print_FiniteElement();

	cout << "---------------Print BoundaryFiniteElement---------------" << endl;
	MyKE.print_BoundaryFiniteElement();

	cout << "---------------Inserting new nodes in the middle of the edges of the KE---------------" << endl;
	MyKE.NewNodes();
	cout << "STL-container nodes:" << endl;
	cout << MyKE.vector_Node() << endl;
	cout << "STL-container finite elements:" << endl;
	cout << MyKE.vector_FiniteElement() << endl;
	cout << "STL-container boundary finite elements:" << endl;
	cout << MyKE.vector_BoundaryFiniteElement() << endl;

	return 0;
}